"""This module contains the data utilities for ``fluke``."""
from __future__ import annotations

import sys
from typing import Optional, Iterable, Any
import warnings

import numpy as np
import rich
import torch
from torch.utils.data import DataLoader, TensorDataset
from numpy.random import (choice, dirichlet, permutation, power, randint,
                          shuffle)
from sklearn.model_selection import train_test_split

sys.path.append(".")
sys.path.append("..")


from .. import DDict  # NOQA

__all__ = [
    'datasets',
    'support',
    'DataContainer',
    'DummyDataContainer',
    'FastDataLoader',
    'DataSplitter'
]


class DataContainer:
    """Container for train and test data.

    Args:
        X_train (torch.Tensor): The training data.
        y_train (torch.Tensor): The training labels.
        X_test (torch.Tensor): The test data.
        y_test (torch.Tensor): The test labels.
        num_classes (int): The number of classes.
        transforms (Optional[callable], optional): The transformation to be applied to the data
          when loaded. Defaults to None.
    """

    def __init__(self,
                 X_train: torch.Tensor,
                 y_train: torch.Tensor,
                 X_test: torch.Tensor,
                 y_test: torch.Tensor,
                 num_classes: int,
                 transforms: Optional[callable] = None):
        if X_train is not None:
            self.train = (X_train, y_train)
            self.test = (X_test, y_test)
            self.num_features = np.prod([i for i in X_train.shape[1:]]).item()
            self.transforms = transforms
        self.num_classes = num_classes


class DummyDataContainer(DataContainer):
    """DataContainer designed for those datasets with a fixed data assignments, e.g., FEMNIST,
    Shakespeare and FCUBE.

    Args:
        clients_tr (Iterable[FastDataLoader]): data loaders for the clients' training set.
        clients_te (Iterable[FastDataLoader]): data loaders for the clients' test set.
        server_data (FastDataLoader): data loader for the server's test set.
    """

    def __init__(self,
                 clients_tr: Iterable[FastDataLoader],
                 clients_te: Iterable[FastDataLoader],
                 server_data: FastDataLoader,
                 num_classes: int):
        super().__init__(None, None, None, None, num_classes=num_classes)
        self.clients_tr = clients_tr
        self.clients_te = clients_te
        self.server_data = server_data


class FastDataLoader:
    """
    A DataLoader-like object for a set of tensors that can be much faster than
    TensorDataset + DataLoader because dataloader grabs individual indices of
    the dataset and calls cat (slow).

    Note:
        This implementation is based on the following discussion:
        https://discuss.pytorch.org/t/dataloader-much-slower-than-manual-batching/27014/6

    Args:
        *tensors (Sequence[torch.Tensor]): tensors to be loaded.
        batch_size (int): batch size.
        shuffle (bool): whether the data should be shuffled.
        transforms (Optional[callable]): the transformation to be applied to the data. Defaults to
          None.
        percentage (float): the percentage of the data to be used.
        skip_singleton (bool): whether to skip batches with a single element. If you have batchnorm
          layers, you might want to set this to ``True``.
        single_batch (bool): whether to return a single batch at each generator iteration.

    Caution:
        When sampling a percentage of the data (i.e., ``percentage < 1``), the data is sampled at
        each epoch. This means that the data varies at each epoch. If you want to keep the data
        constant across epochs, you should sample the data once and pass the sampled
        data to the :class:`FastDataLoader` and set the ``percentage`` parameter to ``1.0``.

    Attributes:
        tensors (Sequence[torch.Tensor]): Tensors of the dataset. Ideally, the first tensor should
          be the input data, and the second tensor should be the labels. However, this is not
          enforced and the user is responsible for ensuring that the tensors are used correctly.
        shuffle (bool): whether the data should be shuffled at each epoch. If ``True``, the data is
          shuffled at each iteration.
        transforms (callable): the transformation to be applied to the data.
        percentage (float): the percentage of the data to be used. If `1.0`, all the data is used.
          Otherwise, the data is sampled according to the given percentage.
        skip_singleton (bool): whether to skip batches with a single element. If you have batchnorm
          layers, you might want to set this to ``True``.
        single_batch (bool): whether to return a single batch at each generator iteration.
        size (int): the size of the dataset according to the percentage of the data to be used.
        max_size (int): the total size (regardless of the sampling percentage) of the dataset.

    Raises:
        AssertionError: if the tensors do not have the same size along the first dimension.
    """

    def __init__(self,
                 *tensors: torch.Tensor,
                 num_labels: int,
                 batch_size: int = 32,
                 shuffle: bool = False,
                 transforms: Optional[callable] = None,
                 percentage: float = 1.0,
                 skip_singleton: bool = True,
                 single_batch: bool = False,
                 **kwargs: dict[str, Any]):
        assert all(t.shape[0] == tensors[0].shape[0] for t in tensors), \
            "All tensors must have the same size along the first dimension."
        self.tensors: Iterable[torch.Tensor] = tensors
        self.num_labels: int = num_labels
        self.max_size = self.tensors[0].shape[0]
        self.percentage: float = percentage
        self.set_sample_size(percentage)
        self.shuffle: bool = shuffle
        self.skip_singleton: bool = skip_singleton
        self.batch_size: int = batch_size if batch_size > 0 else self.size
        self.single_batch: bool = single_batch
        self.transforms: callable = transforms
        self.__i: int = 0

    def asDataLoader(self, **kwargs) -> DataLoader:
        """Convert the ``FastDataLoader`` to a PyTorch DataLoader.

        Args:
            **kwargs: additional arguments for the DataLoader.

        Returns:
            DataLoader: the PyTorch DataLoader.
        """
        return DataLoader(TensorDataset(*self.tensors),
                          batch_size=self.batch_size,
                          shuffle=self.shuffle,
                          drop_last=False,
                          **kwargs)

    def __getitem__(self, index: int) -> tuple:
        """Get the entry at the given index for each tensor.

        Args:
            index (int): the index.

        Raises:
            IndexError: if the index is out of bounds.

        Returns:
            tuple: the entry at the given index for each tensor.
        """
        if index >= self.max_size:
            raise IndexError("Index out of bounds.")
        if self.transforms is not None:
            return (*[self.transforms(t[index])
                      for t in self.tensors[:-1]], self.tensors[-1][index])
        return tuple(t[index] for t in self.tensors)

    def set_sample_size(self, percentage: float) -> int:
        """Set the sample size.

        Args:
            percentage (float): the percentage of the data to be used.

        Returns:
            int: the sample size.
        """
        if percentage > 1.0 or percentage <= 0.0:
            raise ValueError("percentage must be in (0, 1]")
        self.size = max(int(self.tensors[0].shape[0] * percentage), 1)

        if self.size < self.max_size:
            r = torch.randperm(self.max_size)
            self.tensors = [t[r] for t in self.tensors]

        return self.size

    @property
    def batch_size(self) -> int:
        """Get the batch size.

        Returns:
            int: the batch size.
        """
        return self._batch_size

    @batch_size.setter
    def batch_size(self, value: int) -> None:
        """Set the batch size.

        Args:
            value (int): the new batch size.

        Raises:
            ValueError: if the batch size is not positive.
        """
        if value <= 0:
            raise ValueError("batch_size must be > 0")
        self._batch_size = value
        n_batches, remainder = divmod(self.size, self._batch_size)
        if remainder > 0:
            n_batches += 1
        self.n_batches = n_batches

    def __iter__(self):
        if self.shuffle:
            r = torch.randperm(self.size)
            self.tensors = [t[r] for t in self.tensors]
        self.__i = 0
        return self

    def __next__(self) -> tuple:
        if self.single_batch and self.__i > 0:
            raise StopIteration
        if self.__i >= self.size:
            raise StopIteration

        if self.transforms is not None:
            batch = (*[self.transforms(t[self.__i: self.__i+self._batch_size])
                       for t in self.tensors[:-1]],
                     self.tensors[-1][self.__i: self.__i+self._batch_size])
        else:
            batch = tuple(t[self.__i: self.__i+self._batch_size] for t in self.tensors)
        # Useful in case of batch norm layers
        if self.skip_singleton and batch[0].shape[0] == 1:
            raise StopIteration
        self.__i += self._batch_size
        return batch

    def __len__(self) -> int:
        return self.n_batches


class DataSplitter:
    """Utility class for splitting the data across clients."""

    def _safe_train_test_split(self,
                               X: torch.Tensor,
                               y: torch.Tensor,
                               test_size: float,
                               client_id: int | None = None) -> tuple[torch.Tensor, torch.Tensor]:
        try:
            if test_size == 0.0:
                return X, None, y, None
            else:
                return train_test_split(X, y, test_size=test_size, stratify=y)
        except ValueError:
            client_str = f"[Client {client_id}]" if client_id is not None else ""
            rich.print(
                f"[bold red]Warning{client_str}: [/bold red] Stratified split failed. " +
                "Falling back to random split."
            )
            return train_test_split(X, y, test_size=test_size)

    def __init__(self,
                 dataset: DataContainer,
                 distribution: str = "iid",
                 client_split: float = 0.0,
                 sampling_perc: float = 1.0,
                 server_test: bool = True,
                 keep_test: bool = True,
                 server_split: float = 0.0,
                 uniform_test: bool = False,
                 dist_args: DDict = None):
        """Initialize the ``DataSplitter``.

        Args:
            dataset (DataContainer or str): The dataset.
            distribution (str, optional): The data distribution function. Defaults to
              ``"iid"``.
            client_split (float, optional): The size of the client's test set. Defaults to 0.0.
            sampling_perc (float, optional): The percentage of the data to be used. Defaults to 1.0.
            server_test (bool, optional): Whether to keep a server test set. Defaults to True.
            keep_test (bool, optional): Whether to keep the test set provided by the dataset.
              Defaults to ``True``.
            server_split (float, optional): The size of the server's test set. Defaults to 0.0. This
              parameter is used only if ``server_test`` is ``True`` and ``keep_test`` is
              ``False``.
            uniform_test (bool, optional): Whether to distribute the test set in a IID across the
              clients. If ``False``, the test set is distributed according to the distribution
              function. Defaults to ``False``.
            builder_args (DDict, optional): The arguments for the dataset class. Defaults to None.
            dist_args (DDict, optional): The arguments for the distribution function. Defaults to
              ``None``.

        Raises:
            AssertionError: If the parameters are not in the correct range or configuration.
        """
        assert 0 <= client_split <= 1, "client_split must be between 0 and 1."
        assert 0 <= sampling_perc <= 1, "sampling_perc must be between 0 and 1."
        assert 0 <= server_split <= 1, "server_split must be between 0 and 1."
        if not keep_test and server_test and server_split == 0.0:
            raise AssertionError(
                "server_split must be > 0.0 if server_test is True and keep_test is False.")
        if not server_test and client_split == 0.0:
            raise AssertionError("Either client_split > 0 or server_test = True must be true.")

        self.data_container: DataContainer = dataset
        self.distribution: str = distribution
        self.client_split: float = client_split
        self.sampling_perc: float = sampling_perc
        self.keep_test: bool = keep_test
        self.server_test: bool = server_test
        self.server_split: float = server_split
        self.uniform_test: bool = uniform_test
        self.dist_args: DDict = dist_args if dist_args is not None else DDict()

    @property
    def num_classes(self) -> int:
        """Return the number of classes of the dataset.

        Returns:
            int: The number of classes.
        """
        return self.data_container.num_classes

    def assign(self,
               n_clients: int,
               batch_size: int = 32) -> tuple[tuple[FastDataLoader,
                                                    Optional[FastDataLoader]],
                                              FastDataLoader]:
        """Assign the data to the clients and the server according to the configuration.
        Specifically, we can have the following scenarios:

        1. ``server_test = True`` and ``keep_test = True``: The server has a test set that
           corresponds to the test set of the dataset. The clients have a training set and,
           if ``client_split > 0``, a test set.
        2. ``server_test = True`` and ``keep_test = False``: The server has a test set that
           is sampled from the whole dataset (training set and test set are merged). The server's
           sample size is indicated by the ``server_split`` parameter. The clients have a training
           set and, if ``client_split > 0``, a test set.
        3. ``server_test = False`` and ``keep_test = True``: The server does not have a test set.
           The clients have a training set and a test set that corresponds to the test set of the
           dataset distributed uniformly across the clients. In this case the ``client_split`` is
           ignored.
        4. ``server_test = False`` and ``keep_test = False``: The server does not have a test set.
           The clients have a training set and, if ``client_split > 0``, a test set.

        If ``uniform_test = False``, the training and test set are distributed across the clients
        according to the provided distribution. The only exception is done for the test set in
        scenario 3. The test set is IID distributed across clients if ``uniform_test = True``.

        Args:
            n_clients (int): The number of clients.
            batch_size (Optional[int], optional): The batch size. Defaults to 32.

        Returns:
            tuple[tuple[FastDataLoader, Optional[FastDataLoader]], FastDataLoader]:
              The clients' training and testing assignments and the server's testing assignment.
        """
        if isinstance(self.data_container, DummyDataContainer):
            assert n_clients <= len(self.data_container.clients_tr), "n_clients must be <= of " + \
                "the number of clients in the `DummyDataContainer`."
            client_ids = np.sort(np.random.choice(
                len(self.data_container.clients_tr), n_clients, replace=False
            ))
            clients_tr = [self.data_container.clients_tr[i] for i in client_ids]
            clients_te = [self.data_container.clients_te[i] for i in client_ids]
            return (clients_tr, clients_te), self.data_container.server_data

        tranforms = self.data_container.transforms
        if self.server_test and self.keep_test:
            server_X, server_Y = self.data_container.test
            client_X, client_Y = self.data_container.train
            client_Xtr, client_Xte, client_Ytr, client_Yte = self._safe_train_test_split(
                client_X, client_Y, test_size=self.client_split)
        elif not self.keep_test:
            Xtr, ytr = self.data_container.train
            Xte, yte = self.data_container.test
            # Merge and shuffle the data
            X, Y = torch.cat((Xtr, Xte), dim=0), torch.cat((ytr, yte), dim=0)
            idx = torch.randperm(X.size(0))
            X, Y = X[idx], Y[idx]
            # Split the data
            if self.server_test:
                client_X, server_X, client_Y, server_Y = train_test_split(
                    X, Y, test_size=self.server_split)
            else:
                server_X, server_Y = None, None
                client_X, client_Y = X, Y
            client_Xtr, client_Xte, client_Ytr, client_Yte = self._safe_train_test_split(
                client_X, client_Y, test_size=self.client_split)

        else:  # keep_test and not server_test
            server_X, server_Y = None, None
            client_Xtr, client_Ytr = self.data_container.train
            client_Xte, client_Yte = self.data_container.test

        assignments_tr, assignments_te = self._iidness_functions[self.distribution](
            self,
            X_train=client_Xtr,
            y_train=client_Ytr,
            X_test=client_Xte if not self.uniform_test else None,
            y_test=client_Yte if not self.uniform_test else None,
            n=n_clients,
            **self.dist_args
        )

        if client_Xte is not None and self.uniform_test:
            assignments_te, _ = self.iid(client_Xte, client_Yte, None, None, n_clients)

        client_tr_assignments = []
        client_te_assignments = []
        for c in range(n_clients):
            Xtr_client, Ytr_client = client_Xtr[assignments_tr[c]], client_Ytr[assignments_tr[c]]
            client_tr_assignments.append(FastDataLoader(Xtr_client,
                                                        Ytr_client,
                                                        num_labels=self.num_classes,
                                                        batch_size=batch_size,
                                                        shuffle=True,
                                                        transforms=tranforms,
                                                        percentage=self.sampling_perc))
            if assignments_te is not None:
                Xte_client = client_Xte[assignments_te[c]]
                Yte_client = client_Yte[assignments_te[c]]
                client_te_assignments.append(FastDataLoader(Xte_client,
                                                            Yte_client,
                                                            num_labels=self.num_classes,
                                                            batch_size=batch_size,
                                                            shuffle=False,
                                                            percentage=self.sampling_perc))
            else:
                client_te_assignments.append(None)

        server_te = FastDataLoader(server_X, server_Y,
                                   num_labels=self.num_classes,
                                   batch_size=128,
                                   shuffle=False,
                                   percentage=self.sampling_perc) if self.server_test else None
        return (client_tr_assignments, client_te_assignments), server_te

    def iid(self,
            X_train: torch.Tensor,
            y_train: torch.Tensor,
            X_test: Optional[torch.Tensor],
            y_test: Optional[torch.Tensor],
            n: int) -> list[torch.Tensor]:
        """Distribute the examples uniformly across the users.

        Args:
            X_train (torch.Tensor): The training examples.
            y_train (torch.Tensor): The training labels. Not used.
            X_test (torch.Tensor): The test examples.
            y_test (torch.Tensor): The test labels. Not used.
            n (int): The number of clients upon which the examples are distributed.

        Returns:
            list[torch.Tensor]: The examples' ids assignment.
        """
        assert X_train.shape[0] >= n, "# of instances must be > than #clients"
        assert X_test is None or X_test.shape[0] >= n, "# of instances must be > than #clients"

        assigments = []
        for X in (X_train, X_test):
            if X is None:
                assigments.append(None)
                continue
            ex_client = X.shape[0] // n
            idx = np.random.permutation(X.shape[0])
            assigments.append([idx[range(ex_client*i, ex_client*(i+1))] for i in range(n)])
            # Assign the remaining examples one to every client until the examples are finished
            if X.shape[0] % n > 0:
                # assigments[-1] = np.concatenate((assigments[-1], idx[ex_client*n:]))
                for cid, eid in enumerate(range(ex_client*n, X.shape[0])):
                    assigments[-1][cid] = np.append(assigments[-1][cid], idx[eid])

        return assigments[0], assigments[1]

    def quantity_skew(self,
                      X_train: torch.Tensor,
                      y_train: torch.Tensor,  # not used
                      X_test: Optional[torch.Tensor],
                      y_test: Optional[torch.Tensor],  # not used
                      n: int,
                      min_quantity: int = 2,
                      alpha: float = 4.) -> list[torch.Tensor]:
        r"""
        Distribute the examples across the clients according to the following probability density
        function: :math:`P(x; a) = a x^{a-1}`
        where :math:`x` is the id of a client (:math:`x \in [0, n-1]`), and ``a = alpha > 0`` with

        - ``alpha = 1``: examples are equidistributed across clients;
        - ``alpha = 2``: the examples are "linearly" distributed across clients;
        - ``alpha >= 3``: the examples are power law distributed;
        - ``alpha`` :math:`\rightarrow \infty`: all clients but one have ``min_quantity`` examples,
          and the remaining user all the rest.

        Each client is guaranteed to have at least ``min_quantity`` examples.

        Args:
            X_train (torch.Tensor): The training examples.
            y_train (torch.Tensor): The training labels. Not used.
            X_test (torch.Tensor): The test examples.
            y_test (torch.Tensor): The test labels. Not used.
            n (int): The number of clients upon which the examples are distributed.
            min_quantity (int, optional): The minimum number of examples per client. Defaults to 2.
            alpha (float, optional): The skewness parameter. Defaults to 4.

        Returns:
            list[torch.Tensor]: The examples' ids assignment.
        """
        assert min_quantity * \
            n <= X_train.shape[0], "# of training instances must be > than min_quantity*n"
        assert X_test is None or min_quantity * \
            n <= X_test.shape[0], "# of test instances must be > than min_quantity*n"
        assert min_quantity > 0, "min_quantity must be >= 1"

        s = np.array(power(alpha, X_train.shape[0] - min_quantity*n) * n, dtype=int)
        m = np.array([[i] * min_quantity for i in range(n)]).flatten()
        assignment_tr = np.concatenate([s, m])
        shuffle(assignment_tr)

        if X_test is None:
            return [np.where(assignment_tr == i)[0] for i in range(n)], None

        s = np.array(power(alpha, X_test.shape[0] - min_quantity*n) * n, dtype=int)
        m = np.array([[i] * min_quantity for i in range(n)]).flatten()
        assignment_te = np.concatenate([s, m])
        shuffle(assignment_te)

        assignments_tr = [np.where(assignment_tr == i)[0] for i in range(n)]
        assignments_te = [np.where(assignment_te == i)[0] for i in range(n)]
        return assignments_tr, assignments_te

    def label_quantity_skew(self,
                            X_train: torch.Tensor,  # not used
                            y_train: torch.Tensor,
                            X_test: Optional[torch.Tensor],  # not used
                            y_test: Optional[torch.Tensor],
                            n: int,
                            class_per_client: int = 2) -> list[torch.Tensor]:
        """
        This method distribute the data across client according to a specific type of skewness of
        the lables. Specifically:
        suppose each party only has data samples of ``class_per_client`` different labels.
        We first randomly assign ``class_per_client`` different label IDs to each party.
        Then, for the samples of each label, we randomly and equally divide them into the parties
        which own the label. In this way, the number of labels in each party is fixed, and there is
        no overlap between the samples of different parties.
        See: https://arxiv.org/pdf/2102.02079.pdf

        Args:
            X_train (torch.Tensor): The training examples. Not used.
            y_train (torch.Tensor): The training labels.
            X_test (torch.Tensor): The test examples. Not used.
            y_test (torch.Tensor): The test labels.
            n (int): The number of clients upon which the examples are distributed.
            class_per_client (int, optional): The number of classes per client. Defaults to 2.

        Returns:
            list[torch.Tensor]: The examples' ids assignment.
        """
        labels = set(torch.unique(torch.LongTensor(y_train)).numpy())
        assert 0 < class_per_client <= len(labels), "class_per_client must be > 0 and <= #classes"
        assert class_per_client * n >= len(labels), "class_per_client * n must be >= #classes"

        nlbl = [choice(len(labels), class_per_client, replace=False) for _ in range(n)]
        check = set().union(*[set(a) for a in nlbl])
        while len(check) < len(labels):
            missing = labels - check
            for m in missing:
                nlbl[randint(0, n)][randint(0, class_per_client)] = m
            check = set().union(*[set(a) for a in nlbl])
        class_map = {c: [u for u, lbl in enumerate(nlbl) if c in lbl] for c in labels}

        assignments = []
        for y in (y_train, y_test):
            if y is None:
                assignments.append(None)
                continue

            assignment = np.zeros(y.shape[0])
            for lbl, users in class_map.items():
                ids = np.where(y == lbl)[0]
                assignment[ids] = choice(users, len(ids))
            assignments.append([np.where(assignment == i)[0] for i in range(n)])

        return assignments[0], assignments[1]

    def label_dirichlet_skew(self,
                             X_train: torch.Tensor,
                             y_train: torch.Tensor,
                             X_test: Optional[torch.Tensor],
                             y_test: Optional[torch.Tensor],
                             n: int,
                             beta: float = .1,
                             min_ex_class: int = 2,
                             balanced: bool = True) -> list[torch.Tensor]:
        r"""
        The method samples :math:`p_k \sim \text{Dir}_n(\beta)` and allocates a :math:`p_{k,j}`
        proportion of the instances of class :math:`k` to party :math:`j`. Here
        :math:`\text{Dir}(\cdot)` denotes the Dirichlet distribution and beta is a concentration
        parameter :math:`(\beta > 0)`.
        See: https://arxiv.org/pdf/2102.02079.pdf

        Args:
            X_train (torch.Tensor): The training examples.
            y_train (torch.Tensor): The training labels.
            X_test (torch.Tensor): The test examples.
            y_test (torch.Tensor): The test labels.
            n (int): The number of clients upon which the examples are distributed.
            beta (float, optional): The concentration parameter. Defaults to 0.1.
            min_ex_class (int, optional): The minimum number of examples per class. Defaults to 2.
            balanced (bool, optional): Whether to ensure a balanced distribution of the examples.

        Returns:
            list[torch.Tensor]: The examples' ids assignment.
        """
        assert beta > 0, "beta must be > 0"
        assert min_ex_class * \
            n <= X_train.shape[0], "# of training nstances must be >= than min_ex_class * n"
        assert X_test is None or min_ex_class * \
            n <= X_test.shape[0], "# of test instances must be >= than min_ex_class * n"

        labels = list(torch.unique(torch.LongTensor(y_train)).numpy())
        pk = {c: dirichlet([beta]*n) for c in labels}
        for c in labels:
            shuffle(pk[c])

        cid_perm = permutation(n)
        idx_batch = [[[] for _ in range(n)],
                     [[] for _ in range(n)]]
        for iy, y in enumerate([y_train, y_test]):
            if y is None:
                continue

            samples_avg = y.shape[0] / n

            shuffle(labels)
            for c in labels:
                # get the ids of the examples of class c
                ids = np.where(y == c)[0]
                shuffle(ids)

                # Assign the minimum number of examples per class
                idx_batch[iy] = [
                    idx_j + ids[i * min_ex_class: (i + 1) * min_ex_class].tolist()
                    for i, idx_j in enumerate(idx_batch[iy])
                ]
                ids = ids[n * min_ex_class:]

                # compute the proportions
                proportions = np.array([p * (len(idx_j) < samples_avg)
                                        for p, idx_j in zip(pk[c], idx_batch[iy])])
                proportions = proportions / proportions.sum()

                # assign the examples to the clients
                proportions = (np.cumsum(proportions) * len(ids)).astype(int)[:-1]
                idx_batch[iy] = [idx_j + idx.tolist()
                                 for idx_j, idx in zip(idx_batch[iy],
                                                       np.split(ids, proportions))]

        trials = 0
        if balanced:
            for iy, y in enumerate([y_train, y_test]):
                if y is None:
                    continue
                samples_avg = np.ceil(y.shape[0] / n)

                to_reduce = [i for i in range(len(idx_batch[iy]))
                             if len(idx_batch[iy][i]) - int(samples_avg) > 0]

                to_fill = [i for i in range(n) if i not in to_reduce]

                while to_reduce and trials < 10:
                    i = to_reduce.pop(0)
                    samples = idx_batch[iy][i]
                    surplus_ratio = 1. - samples_avg / len(samples)
                    try:
                        s_keep, s_spare, _, _ = train_test_split(samples, y[samples],
                                                                 test_size=surplus_ratio,
                                                                 stratify=y[samples])
                    except ValueError:
                        s_keep, s_spare, _, _ = train_test_split(samples, y[samples],
                                                                 test_size=surplus_ratio)
                    idx_batch[iy][i] = s_keep

                    step = int(np.ceil(len(s_spare) / len(to_fill)))
                    shuffle(s_spare)
                    new_reduce = []
                    no_fill = []
                    for jj, j in enumerate(to_fill):
                        idx_batch[iy][j] += s_spare[step*jj: step*(jj+1)]
                        if len(idx_batch[iy][j]) - int(samples_avg) > 0:
                            new_reduce.append(j)
                            no_fill.append(j)
                        elif len(idx_batch[iy][j]) - int(samples_avg) == 0:
                            no_fill.append(j)

                    to_reduce += new_reduce
                    for j in no_fill:
                        to_fill.remove(j)

                    if len(idx_batch[iy][i]) < samples_avg:
                        to_fill.append(i)

                    trials += 1

        if trials > 10:
            warnings.warn(
                "Reached maximum number of trials (10) while trying to balanced the dataset \
                    distribution")
        # change idx_batch according to cid_perm
        idx_batch = [[idx_batch[i][cid_perm[j]] for j in range(n)] for i in range(2)]
        return idx_batch[0], idx_batch[1] if y_test is not None else None

    def label_pathological_skew(self,
                                X_train: torch.Tensor,  # not used
                                y_train: torch.Tensor,
                                X_test: Optional[torch.Tensor],  # not used
                                y_test: Optional[torch.Tensor],
                                n: int,
                                shards_per_client: int = 2) -> list[torch.Tensor]:
        """
        The method first sort the data by label, divide it into ``n * shards_per_client`` shards,
        and assign each of ``n`` clients ``shards_per_client`` shards. This is a pathological
        non-IID partition of the data, as most clients will only have examples of a limited number
        of classes.
        See: http://proceedings.mlr.press/v54/mcmahan17a/mcmahan17a.pdf

        Args:
            X_train (torch.Tensor): The training examples. Not used.
            y_train (torch.Tensor): The training labels.
            X_test (torch.Tensor): The test examples. Not used.
            y_test (torch.Tensor): The test labels.
            n (int): The number of clients upon which the examples are distributed.
            shards_per_client (int, optional): The number of shards per client. Defaults to 2.

        Returns:
            list[torch.Tensor]: The examples' ids assignment.
        """
        n_shards = int(shards_per_client * n)
        sorted_ids_tr = np.argsort(y_train.numpy())
        shard_size_tr = int(np.ceil(len(y_train) / n_shards))
        assignments_tr = np.zeros(y_train.shape[0])

        assignments_te = None
        if y_test is not None:
            sorted_ids_te = np.argsort(y_test.numpy())
            shard_size_te = int(np.ceil(len(y_test) / n_shards))
            assignments_te = np.zeros(y_test.shape[0])

        perm = permutation(n_shards)
        j = 0
        for i in range(n):
            for _ in range(shards_per_client):
                left = perm[j] * shard_size_tr
                right = min((perm[j]+1) * shard_size_tr, len(y_train))
                assignments_tr[sorted_ids_tr[left:right]] = i

                if y_test is not None:
                    left = perm[j] * shard_size_te
                    right = min((perm[j]+1) * shard_size_te, len(y_test))
                    assignments_te[sorted_ids_te[left:right]] = i

                j += 1
        assignments_tr = [np.where(assignments_tr == i)[0] for i in range(n)]
        if y_test is not None:
            assignments_te = [np.where(assignments_te == i)[0] for i in range(n)]
        return assignments_tr, assignments_te

    _iidness_functions = {
        "iid": iid,
        "qnt": quantity_skew,
        # "classwise_qnt": classwise_quantity_skew,
        "lbl_qnt": label_quantity_skew,
        "dir": label_dirichlet_skew,
        "pathological": label_pathological_skew,
        # "covariate": covariate_shift
    }



"""
This module contains the datasets for loading the supported datasets.
"""
from __future__ import annotations

import json
import os
import string
import sys
from typing import Optional

import numpy as np
import torch
from datasets import load_dataset
from numpy.random import permutation
from rich.progress import track
from torchvision import datasets
from torchvision.datasets import VisionDataset
from torchvision.transforms import Lambda, ToTensor

sys.path.append(".")
sys.path.append("..")

from ..utils import get_class_from_qualified_name  # NOQA
from . import DataContainer, DummyDataContainer, FastDataLoader, support  # NOQA

__all__ = [
    "Datasets"
]


def _apply_transforms(dataset: VisionDataset, transforms: Optional[callable]) -> VisionDataset:
    if transforms is not None:
        new_data = []
        for i in range(len(dataset)):
            new_data.append(transforms(dataset[i][0]))
        dataset.data = torch.stack(new_data)
        # dataset.data = transforms(dataset.data)
    dataset.data = torch.Tensor(dataset.data)
    dataset.targets = torch.LongTensor(dataset.targets)
    return dataset


class Datasets:
    """Static class for loading datasets.
    Datasets are downloaded (if needed) into the ``path`` folder. The supported datasets are:
    ``MNIST``, ``MNISTM``, ``SVHN``, ``FEMNIST``, ``EMNIST``, ``CIFAR10``, ``CIFAR100``,
    ``Tiny Imagenet``, ``Shakespeare``, ``Fashion MNIST``, and ``CINIC10``.
    Each dataset but ``femnist`` and ``shakespeare`` can be transformed using the ``transforms``
    argument. Each dataset is returned as a :class:`fluke.data.DataContainer` object.

    .. important::
        ``onthefly_transforms`` are transformations that are applied on-the-fly to the data
        through the data loader. This is useful when the transformations are stochastic and
        should be applied at each iteration. These transformations cannot be configured through
        the configuration file.
    """

    @classmethod
    def get(cls, name: str, **kwargs) -> DataContainer | tuple:
        """Get a dataset by name initialized with the provided arguments.
        Supported datasets are: ``mnist``, ``mnistm``, ``svhn``, ``femnist``, ``emnist``,
        ``cifar10``, ``cifar100``, ``tiny_imagenet``, ``shakespeare``, ``fashion_mnist``, and
        ``cinic10``. If `name` is not in the supported datasets, it is assumed to be a fully
        qualified name of a custom dataset function (``callable[..., DataContainer]``).

        Args:
            name (str): The name of the dataset to load or the fully qualified name of a custom
              dataset function.
            **kwargs: Additional arguments to pass to construct the dataset.

        Returns:
            DataContainer: The ``DataContainer`` object containing the dataset.

        Raises:
            ValueError: If the dataset is not supported or the name is wrong.
        """
        if name not in Datasets._DATASET_MAP:
            try:
                data_fun = get_class_from_qualified_name(name)
                return data_fun(**kwargs)
            except (ModuleNotFoundError, TypeError, ValueError) as e:
                if "." in name:
                    raise e
                else:
                    raise ValueError(f"Dataset {name} not found. The supported datasets are: " +
                                     ", ".join(Datasets._DATASET_MAP.keys()) + ".")

        return Datasets._DATASET_MAP[name](**kwargs)

    @classmethod
    def MNIST(cls,
              path: str = "../data",
              transforms: Optional[callable] = None,
              onthefly_transforms: Optional[callable] = None,
              channel_dim: bool = False) -> DataContainer:
        """Load the MNIST dataset.
        The dataset is split into training and testing sets according to the default split of the
        :class:`torchvision.datasets.MNIST` class. If no transformations are provided, the data is
        normalized to the range [0, 1].
        An example of the dataset is a 28x28 image, i.e., a tensor of shape (28, 28).
        The dataset has 10 classes, corresponding to the digits 0-9.

        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"../data"``.
            transforms (callable, optional): The transformations to apply to the data. Defaults to
              ``None``.
            onthefly_transforms (callable, optional): The transformations to apply on-the-fly to the
              data through the data loader. Defaults to ``None``.
            channel_dim (bool, optional): Whether to add a channel dimension to the data, i.e., the
              shape of the an example becomes (1, 28, 28). Defaults to ``False``.

        Returns:
            DataContainer: The MNIST dataset.
        """
        train_data = datasets.MNIST(
            root=path,
            train=True,
            download=True
        )

        test_data = datasets.MNIST(
            root=path,
            train=False,
            download=True
        )

        train_data = _apply_transforms(train_data, transforms)
        test_data = _apply_transforms(test_data, transforms)

        if transforms is None:
            train_data.data = torch.Tensor(train_data.data / 255.)
            test_data.data = torch.Tensor(test_data.data / 255.)

        return DataContainer(train_data.data if not channel_dim else train_data.data[:, None, :, :],
                             train_data.targets,
                             test_data.data if not channel_dim else test_data.data[:, None, :, :],
                             test_data.targets,
                             10,
                             onthefly_transforms)

    @classmethod
    def MNISTM(cls,
               path: str = "../data",
               transforms: Optional[callable] = None,
               onthefly_transforms: Optional[callable] = None) -> DataContainer:
        """Load the MNIST-M dataset. MNIST-M is a dataset where the MNIST digits are placed on
        random color patches. The dataset is split into training and testing sets according to the
        default split of the data at https://github.com/liyxi/mnist-m/releases/download/data/.
        If no transformations are provided, the data is normalized to the range [0, 1].
        The dataset has 10 classes, corresponding to the digits 0-9.

        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"../data"``.
            transforms (callable, optional): The transformations to apply to the data. Defaults to
              ``None``.
            onthefly_transforms (callable, optional): The transformations to apply on-the-fly to the
              data through the data loader. Defaults to ``None``.

        Returns:
            DataContainer: The MNIST-M dataset.
        """
        train_data = support.MNISTM(
            root=path,
            train=True,
            download=True,
        )

        test_data = support.MNISTM(
            root=path,
            train=False,
            download=True
        )

        train_data = _apply_transforms(train_data, transforms)
        test_data = _apply_transforms(test_data, transforms)

        if transforms is None:
            train_data.data = train_data.data / 255.
            test_data.data = test_data.data / 255.

            train_data.data = torch.movedim(train_data.data, 3, 1)
            test_data.data = torch.movedim(test_data.data, 3, 1)

        return DataContainer(train_data.data,
                             train_data.targets,
                             test_data.data,
                             test_data.targets,
                             10,
                             onthefly_transforms)

    @classmethod
    def EMNIST(cls,
               path: str = "../data",
               transforms: Optional[callable] = None,
               onthefly_transforms: Optional[callable] = None,
               channel_dim: bool = False) -> DataContainer:
        """Load the Extended MNIST (EMNIST) dataset. The dataset is split into training and testing
        sets according to the default split of the data at
        https://www.westernsydney.edu.au/bens/home/reproducible_research/emnist as provided by
        the :class:`torchvision.datasets.EMNIST` class.
        If no transformations are provided, the data is normalized to the range [0, 1]. The dataset
        has 47 classes, corresponding to the digits 0-9 and the uppercase and lowercase letters.

        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"../data"``.
            transforms (callable, optional): The transformations to apply to the data. Defaults to
              ``None``.
            onthefly_transforms (callable, optional): The transformations to apply on-the-fly to the
              data through the data loader. Defaults to ``None``.

        Returns:
            DataContainer: The EMNIST dataset.
        """

        train_data = datasets.EMNIST(
            root=path,
            split="balanced",
            train=True,
            download=True
        )

        test_data = datasets.EMNIST(
            root=path,
            split="balanced",
            train=False,
            download=True
        )

        train_data = _apply_transforms(train_data, transforms)
        test_data = _apply_transforms(test_data, transforms)

        if transforms is None:
            train_data.data = train_data.data / 255.
            test_data.data = test_data.data / 255.

        return DataContainer(train_data.data if not channel_dim else train_data.data[:, None, :, :],
                             train_data.targets,
                             test_data.data if not channel_dim else test_data.data[:, None, :, :],
                             test_data.targets,
                             47,
                             onthefly_transforms)

    @classmethod
    def SVHN(cls,
             path: str = "../data",
             transforms: Optional[callable] = None,
             onthefly_transforms: Optional[callable] = None) -> DataContainer:
        """
        Load the Street View House Numbers (SVHN) dataset. The dataset is split into training and
        testing sets according to the default split of the :class:`torchvision.datasets.SVHN` class.
        If no transformations are provided, the data is normalized to the range [0, 1]. The dataset
        contains 10 classes, corresponding to the digits 0-9.

        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"../data"``.
            transforms (callable, optional): The transformations to apply to the data. Defaults to
              ``None``.
            onthefly_transforms (callable, optional): The transformations to apply on-the-fly to the
              data through the data loader. Defaults to ``None``.

        Returns:
            DataContainer: The SVHN dataset.
        """
        train_data = datasets.SVHN(
            root=path,
            split="train",
            download=True
        )

        test_data = datasets.SVHN(
            root=path,
            split="test",
            download=True
        )

        # Force targets to be named "targets" instead of "labels" for compatibility
        setattr(train_data, "targets", train_data.labels)
        setattr(test_data, "targets", test_data.labels)

        train_data = _apply_transforms(train_data, transforms)
        test_data = _apply_transforms(test_data, transforms)

        if transforms is None:
            train_data.data = torch.Tensor(train_data.data / 255.)
            test_data.data = torch.Tensor(test_data.data / 255.)

        return DataContainer(train_data.data,
                             train_data.targets,
                             test_data.data,
                             test_data.targets,
                             10,
                             onthefly_transforms)

    @classmethod
    def CIFAR10(cls,
                path: str = "../data",
                transforms: Optional[callable] = None,
                onthefly_transforms: Optional[callable] = None) -> DataContainer:
        """
        Load the CIFAR-10 dataset. The dataset is split into training and testing sets according to
        the default split of the :class:`torchvision.datasets.CIFAR10` class.
        If no transformations are provided, the data is normalized to the range [0, 1]. The dataset
        contains 10 classes, corresponding to the following classes: ``airplane``, ``automobile``,
        ``bird``, ``cat``, ``deer``, ``dog``, ``frog``, ``horse``, ``ship``, and ``truck``.
        The images shape is (3, 32, 32).


        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"../data"``.
            transforms (callable, optional): The transformations to apply to the data. Defaults to
              ``None``.
            onthefly_transforms (callable, optional): The transformations to apply on-the-fly to the
              data through the data loader. Defaults to ``None``.

        Returns:
            DataContainer: The CIFAR-10 dataset.
        """
        train_data = datasets.CIFAR10(
            root=path,
            train=True,
            download=True
        )

        test_data = datasets.CIFAR10(
            root=path,
            train=False,
            download=True
        )

        train_data = _apply_transforms(train_data, transforms)
        test_data = _apply_transforms(test_data, transforms)

        if transforms is None:
            train_data.data = train_data.data / 255.
            test_data.data = test_data.data / 255.

            train_data.data = torch.movedim(train_data.data, 3, 1)
            test_data.data = torch.movedim(test_data.data, 3, 1)

        return DataContainer(train_data.data,
                             train_data.targets,
                             test_data.data,
                             test_data.targets,
                             10,
                             onthefly_transforms)

    @classmethod
    def CINIC10(cls,
                path: str = "../data",
                transforms: Optional[callable] = None,
                onthefly_transforms: Optional[callable] = None) -> DataContainer:
        """
        Load the CINIC-10 dataset. `CINIC-10 <http://dx.doi.org/10.7488/ds/2448>`_ is an
        augmented extension of CIFAR-10. It contains the images from CIFAR-10
        (60,000 images, 32x32 RGB pixels) and a selection of ImageNet database images
        (210,000 images downsampled to 32x32). It was compiled as a 'bridge' between CIFAR-10 and
        ImageNet, for benchmarking machine learning applications. It is split into three equal
        subsets - train, validation, and test - each of which contain 90,000 images.

        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"../data"``.
            transforms (callable, optional): The transformations to apply to the data. Defaults to
              ``None``.
            onthefly_transforms (callable, optional): The transformations to apply on-the-fly to the
              data through the data loader. Defaults to ``None``.

        Returns:
            DataContainer: The CINIC-10 dataset.
        """

        train_data = support.CINIC10(
            root=path,
            split="train",
            download=True
        )

        test_data = support.CINIC10(
            root=path,
            split="test",
            download=True
        )

        train_data = _apply_transforms(train_data, transforms)
        test_data = _apply_transforms(test_data, transforms)

        return DataContainer(train_data.data,
                             train_data.targets,
                             test_data.data,
                             test_data.targets,
                             10,
                             onthefly_transforms)

    @classmethod
    def CIFAR100(cls,
                 path: str = "../data",
                 transforms: Optional[callable] = None,
                 onthefly_transforms: Optional[callable] = None) -> DataContainer:
        """
        Load the CIFAR-100 dataset. The dataset is split into training and testing sets according to
        the default split of the :class:`torchvision.datasets.CIFAR100` class.
        If no transformations are provided, the data is normalized to the range [0, 1]. The dataset
        contains 100 classes, corresponding to different type of objects. The images shape is
        (3, 32, 32).

        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"../data"``.
            transforms (callable, optional): The transformations to apply to the data. Defaults to
              ``None``.
            onthefly_transforms (callable, optional): The transformations to apply on-the-fly to the
              data through the data loader. Defaults to ``None``.

        Returns:
            DataContainer: The CIFAR-100 dataset.
        """
        train_data = datasets.CIFAR100(
            root=path,
            train=True,
            download=True
        )

        test_data = datasets.CIFAR100(
            root=path,
            train=False,
            download=True
        )

        train_data = _apply_transforms(train_data, transforms)
        test_data = _apply_transforms(test_data, transforms)

        if transforms is None:
            train_data.data = train_data.data / 255.
            test_data.data = test_data.data / 255.

            train_data.data = torch.movedim(train_data.data, 3, 1)
            test_data.data = torch.movedim(test_data.data, 3, 1)

        return DataContainer(train_data.data,
                             train_data.targets,
                             test_data.data,
                             test_data.targets,
                             100,
                             onthefly_transforms)

    @classmethod
    def FASHION_MNIST(cls,
                      path: str = "../data",
                      transforms: Optional[callable] = None,
                      onthefly_transforms: Optional[callable] = None) -> DataContainer:
        """
        Load the Fashion MNIST dataset. The dataset is split into training and testing sets
        according to the default split of the :class:`torchvision.datasets.FashionMNIST` class.
        The dataset contains 10 classes, corresponding to different types of clothing.
        The images shape is (28, 28).

        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"../data"``.
            transforms (callable, optional): The transformations to apply to the data. Defaults to
              ``None``.
            onthefly_transforms (callable, optional): The transformations to apply on-the-fly to the
              data through the data loader. Defaults to ``None``.

        Returns:
            DataContainer: The CIFAR-100 dataset.
        """
        train_data = datasets.FashionMNIST(
            root=path,
            train=True,
            download=True
        )

        test_data = datasets.FashionMNIST(
            root=path,
            train=False,
            download=True
        )

        train_data = _apply_transforms(train_data, transforms)
        test_data = _apply_transforms(test_data, transforms)

        return DataContainer(train_data.data,
                             train_data.targets,
                             test_data.data,
                             test_data.targets,
                             10,
                             onthefly_transforms)

    @classmethod
    def TINY_IMAGENET(cls,
                      path: str = "../data",
                      transforms: Optional[callable] = None,
                      onthefly_transforms: Optional[callable] = None) -> DataContainer:
        """
        Load the Tiny-ImageNet dataset.
        This version of the dataset is the one offered by the
        `Hugging Face <https://huggingface.co/datasets/zh-plus/tiny-imagenet>`_. The dataset is
        split into training and testing sets according to the default split of the data.
        The dataset contains 200 classes, corresponding to different types of objects.
        The images shape is (3, 64, 64).

        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"../data"``.
            transforms (callable, optional): The transformations to apply to the data. Defaults to
              ``None``.
            onthefly_transforms (callable, optional): The transformations to apply on-the-fly to the
              data through the data loader. Defaults to ``None``.

        Returns:
            DataContainer: The Tiny-ImageNet dataset.
        """
        tiny_imagenet_train = load_dataset('Maysee/tiny-imagenet',
                                           split='train',
                                           cache_dir=path)

        tiny_imagenet_test = load_dataset('Maysee/tiny-imagenet',
                                          split='valid',
                                          cache_dir=path)

        X_train, y_train = [], []
        X_test, y_test = [], []

        fix_bw_trans = Lambda(lambda x: x.repeat(1, 3, 1, 1))

        for image in track(tiny_imagenet_train, "Loading Tiny ImageNet train data..."):
            y = image['label']
            image = image['image']
            if transforms is not None:
                x = transforms(image)
            if not isinstance(image, torch.Tensor):
                x = ToTensor()(image).unsqueeze(0)
            else:
                x = image.unsqueeze(0)
            if x.shape != torch.Size([1, 3, 64, 64]):
                x = fix_bw_trans(x)
            X_train.append(x)
            y_train.append(y)

        for image in track(tiny_imagenet_test, "Loading Tiny ImageNet test data..."):
            y = image['label']
            image = image['image']
            if transforms is not None:
                image = transforms(image)
            if not isinstance(image, torch.Tensor):
                x = ToTensor()(image).unsqueeze(0)
            else:
                x = image.unsqueeze(0)
            if x.shape != torch.Size([1, 3, 64, 64]):
                x = fix_bw_trans(x)
            X_test.append(x)
            y_test.append(y)

        train_data = torch.vstack(X_train)
        test_data = torch.vstack(X_test)

        # if transforms is not None:
        #     train_data = transforms(train_data)
        #     test_data = transforms(test_data)

        idxperm = torch.randperm(train_data.shape[0])
        train_data = train_data[idxperm]
        y_train = torch.LongTensor(y_train)[idxperm]
        y_test = torch.LongTensor(y_test)

        return DataContainer(train_data,
                             y_train,
                             test_data,
                             y_test,
                             200,
                             onthefly_transforms)

    @classmethod
    def FEMNIST(cls,
                path: str = "./data",
                batch_size: int = 10,
                filter: str = "all",
                onthefly_transforms: Optional[callable] = None) -> DataContainer:
        """
        Load the Federated EMNIST (FEMNIST) dataset.
        This dataset is the one offered by the `Leaf project <https://leaf.cmu.edu/>`_.
        FEMNIST contains images of handwritten digits of size 28 by 28 pixels (with option to make
        them all 128 by 128 pixels) taken from 3500 users.
        The dataset has 62 classes corresponding to different characters.
        The label-class correspondence is as follows:

        .. code-block:: rst

            classes: 0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz
            labels : 01234567890123456789012345678901234567890123456789012345678901
                               10        20        30        40        50        60

        Important:
            Differently from the other datasets (but :meth:`SHAKESPEARE`), the FEMNIST dataset can
            not be downloaded directly from ``fluke`` but it must be downloaded from the
            `Leaf project <https://leaf.cmu.edu/>`_ and stored in the ``path`` folder.
            The datasets must also be created according to the instructions provided by the Leaf
            project. The expected folder structure is:

            .. code-block:: bash

                path
                ├── FEMNIST
                │   ├── train
                │   │   ├── user_data_0.json
                │   │   ├── user_data_1.json
                │   │   └── ...
                │   └── test
                │       ├── user_data_0.json
                │       ├── user_data_1.json
                │       └── ...

            where in each ``user_data_X.json`` file there is a dictionary with the keys
            ``user_data`` containing the data of the user.

        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"./data"``.
            batch_size (int, optional): The batch size. Defaults to ``10``.
            filter (str, optional): The filter for the selection of a specific portion of the
                dataset. The options are: ``all``, ``uppercase``, ``lowercase``, and ``digits``.
                Defaults to ``"all"``.

        Returns:
            tuple: A tuple containing the training and testing data loaders for the clients. The
                server data loader is ``None``.
        """
        def _filter_femnist(udata, filter):
            # classes: 0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz
            # labels : 01234567890123456789012345678901234567890123456789012345678901
            if filter == "all":
                return udata, 62
            elif filter == "uppercase":
                udata["x"] = [x for x, y in zip(udata["x"], udata["y"]) if y < 36 and y > 9]
                udata["y"] = [y - 10 for y in udata["y"] if y < 36 and y > 9]
                num_classes = 26
            elif filter == "lowercase":
                udata["x"] = [x for x, y in zip(udata["x"], udata["y"]) if y > 35]
                udata["y"] = [y - 36 for y in udata["y"] if y > 35]
                num_classes = 26
            elif filter == "digits":
                udata["x"] = [x for x, y in zip(udata["x"], udata["y"]) if y < 10]
                udata["y"] = [y for y in udata["y"] if y < 10]
                num_classes = 10
            else:
                raise ValueError(f"Invalid filter: {filter}")
            return udata, num_classes

        femnist_path = os.path.join(path, "FEMNIST")
        train_dir = os.path.join(femnist_path, 'train')
        test_dir = os.path.join(femnist_path, 'test')

        assert os.path.exists(femnist_path), f"FEMNIST data ({femnist_path}) not found."
        assert os.path.exists(train_dir), f"FEMNIST train data ({train_dir}') not found."
        assert os.path.exists(test_dir), f"FEMNIST test data ({test_dir}') not found."

        # TRAINING
        files = os.listdir(train_dir)
        dict_train = {}
        for file in track(files, "Loading FEMNIST train data..."):
            with open(os.path.join(train_dir, file)) as f:
                data = json.load(f)
            dict_train.update(data["user_data"])

        # TEST
        files = os.listdir(test_dir)
        dict_test = {}
        for file in track(files, "Loading FEMNIST test data..."):
            with open(os.path.join(test_dir, file)) as f:
                data = json.load(f)
            dict_test.update(data["user_data"])

        client_tr_assignments = []
        for k in track(sorted(dict_train), "Creating training data loader..."):
            udata = dict_train[k]
            udata, num_classes = _filter_femnist(udata, filter)
            Xtr_client = torch.FloatTensor(udata["x"]).reshape(-1, 1, 28, 28)
            ytr_client = torch.LongTensor(udata["y"])
            client_tr_assignments.append(
                FastDataLoader(
                    Xtr_client,
                    ytr_client,
                    num_labels=num_classes,
                    batch_size=batch_size,
                    shuffle=True,
                    transforms=onthefly_transforms,
                    percentage=1.0
                )
            )

        client_te_assignments = []
        for k in track(sorted(dict_train), "Creating testing data loader..."):
            udata = dict_test[k]
            udata, _ = _filter_femnist(udata, filter)
            Xte_client = torch.FloatTensor(udata["x"]).reshape(-1, 1, 28, 28)
            yte_client = torch.LongTensor(udata["y"])
            client_te_assignments.append(
                FastDataLoader(
                    Xte_client,
                    yte_client,
                    num_labels=num_classes,
                    batch_size=64,
                    shuffle=True,
                    transforms=onthefly_transforms,
                    percentage=1.0
                )
            )

        perm = permutation(len(client_tr_assignments))
        client_tr_assignments = [client_tr_assignments[i] for i in perm]
        client_te_assignments = [client_te_assignments[i] for i in perm]
        return DummyDataContainer(client_tr_assignments, client_te_assignments, None, 62)

    @classmethod
    def SHAKESPEARE(cls,
                    path: str = "./data",
                    batch_size: int = 10,
                    onthefly_transforms: Optional[callable] = None) -> tuple:
        """Load the Federated Shakespeare dataset.
        This dataset is the one offered by the `Leaf project <https://leaf.cmu.edu/>`_.
        Shakespeare is a text dataset containing dialogues from Shakespeare's plays.
        Dialogues are taken from 660 users and the task is to predict the next character in a
        dialogue (which is solved as a classification problem with 100 classes).

        Important:
            Differently from the other datasets (but :meth:`FEMNIST`), the ``SHAKESPEARE`` dataset
            can not be downloaded directly from ``fluke`` but it must be downloaded from the
            `Leaf project <https://leaf.cmu.edu/>`_ and stored in the ``path`` folder.
            The datasets must also be created according to the instructions provided by the Leaf
            project. The expected folder structure is:

            .. code-block:: bash

                path
                ├── shakespeare
                │   ├── train
                │   │   ├── user_data_0.json
                │   │   ├── user_data_1.json
                │   │   └── ...
                │   └── test
                │       ├── user_data_0.json
                │       ├── user_data_1.json
                │       └── ...

            where in each ``user_data_X.json`` file there is a dictionary with the keys
            ``user_data`` containing the data of the user.

        Args:
            path (str, optional): The path where the dataset is stored. Defaults to ``"./data"``.
            batch_size (int, optional): The batch size. Defaults to ``10``.

        Returns:
            tuple: A tuple containing the training and testing data loaders for the clients. The
                server data loader is ``None``.
        """
        shake_path = os.path.join(path, "shakespeare")
        train_dir = os.path.join(shake_path, 'train')
        test_dir = os.path.join(shake_path, 'test')

        assert os.path.exists(shake_path), f"shakespeare data ({shake_path}) not found."
        assert os.path.exists(train_dir), f"shakespeare train data ({train_dir}') not found."
        assert os.path.exists(test_dir), f"shakespeare test data ({test_dir}') not found."

        all_chr = string.printable
        chr_map = {c: i for i, c in enumerate(all_chr)}

        # TRAINING
        files = os.listdir(train_dir)
        dict_train = {}
        for file in track(files, "Loading Shakespeare train data..."):
            with open(os.path.join(train_dir, file)) as f:
                data = json.load(f)
            dict_train.update(data["user_data"])

        # TEST
        files = os.listdir(test_dir)
        dict_test = {}
        for file in track(files, "Loading Shakespeare test data..."):
            with open(os.path.join(test_dir, file)) as f:
                data = json.load(f)
            dict_test.update(data["user_data"])

        client_tr_assignments = []
        for k in track(sorted(dict_train), "Creating training data loader..."):
            udata = dict_train[k]
            inputs, targets = udata['x'], udata['y']
            for idx in range(len(inputs)):
                inputs[idx] = [chr_map[char] for char in inputs[idx]]
            for idx in range(len(targets)):
                targets[idx] = chr_map[targets[idx]]

            Xtr_client = torch.LongTensor(inputs)
            ytr_client = torch.LongTensor(targets)
            client_tr_assignments.append(
                FastDataLoader(
                    Xtr_client,
                    ytr_client,
                    num_labels=100,
                    batch_size=batch_size,
                    shuffle=True,
                    transforms=onthefly_transforms,
                    percentage=1.0
                )
            )

        client_te_assignments = []
        for k in track(sorted(dict_train), "Creating test data loader..."):
            udata = dict_test[k]
            inputs, targets = udata['x'], udata['y']
            for idx in range(len(inputs)):
                inputs[idx] = [chr_map[char] for char in inputs[idx]]
            for idx in range(len(targets)):
                targets[idx] = chr_map[targets[idx]]

            Xte_client = torch.LongTensor(inputs)
            yte_client = torch.LongTensor(targets)
            client_te_assignments.append(
                FastDataLoader(
                    Xte_client,
                    yte_client,
                    num_labels=100,
                    batch_size=batch_size,
                    shuffle=True,
                    transforms=onthefly_transforms,
                    percentage=1.0
                )
            )

        perm = permutation(len(client_tr_assignments))
        client_tr_assignments = [client_tr_assignments[i] for i in perm]
        client_te_assignments = [client_te_assignments[i] for i in perm]
        return DummyDataContainer(client_tr_assignments, client_te_assignments, None, 100)

    @classmethod
    def FCUBE(cls,
              path: str = "./data",
              batch_size: int = 10,
              n_points: int = 1000,
              test_size: int = 0.1,
              dimensions: int = 3) -> tuple:
        """This class creates the dataset FCUBE as described in the paper
        https://arxiv.org/pdf/2102.02079. This implementation generalizes for $n$ dimensions.
        The number of clients depends on the number of dimensions, i.e., the dataset will be divided
        in a number of partitions that is equal to $2^{d-1}$ where $d$ is the number of dimensions.

        Warning:
            This procedure may become very slow for large value of `dimensions`, e.g.,
            `dimensions > 8`.

        Args:
            path (str, optional): The path where to save or load the dataset. Defaults to "./data".
            batch_size (int, optional): The batch size. Defaults to 10.
            n_points (int, optional): The total number of points to generate. Defaults to 1000.
            test_size (int, optional): The percentage of points to include in the test sets for both
              the clients and the server. Defaults to 0.1.
            dimensions (int, optional): The number of dimensions of the points. Defaults to 3.

        Returns:
            tuple: _description_
        """

        root = os.path.join(path, "FCUBE")
        os.makedirs(root, exist_ok=True)
        file_path = os.path.join(root, "fcube.pth")
        if os.path.exists(file_path):
            data = torch.load(file_path, weights_only=False)
            X, y = data["X"], data["y"]
        else:
            X = torch.rand((n_points, dimensions)) * 2 - 1
            y = torch.LongTensor([0 if x[0] <= 0 else 1 for x in X])
            torch.save({"X": X, "y": y}, file_path)

        n_clients = 2**(dimensions-1)
        test_percent = int(n_points * test_size)
        server_te = FastDataLoader(
            X[-test_percent:],
            y[-test_percent:],
            num_labels=2,
            batch_size=batch_size,
            shuffle=False,
            percentage=1.0
        )

        def to_bin(v):
            return sum(map(lambda x: x[1] << x[0], enumerate(v)))

        Xbin = [to_bin(x).item() for x in (X[:-test_percent] > 0)]

        max_id = 2**dimensions
        bin_to_idx = {i: [] for i in range(max_id)}
        for i, b in enumerate(Xbin):
            bin_to_idx[b].append(i)

        clients_per_bin = [0] * len(bin_to_idx)
        for i in range(n_clients):
            pos_id = i % max_id
            neg_id = max_id - pos_id - 1

            clients_per_bin[pos_id] += 1
            clients_per_bin[neg_id] += 1

        point_per_bin = [int(np.ceil(len(bin_to_idx[i]) / clients_per_bin[i]))
                         for i in range(len(bin_to_idx))]
        cnt_bin = [0] * len(bin_to_idx)

        client_tr_assignments = []
        client_te_assignments = []
        for i in range(n_clients):
            pos_id = i % max_id
            neg_id = max_id - pos_id - 1

            step = point_per_bin[pos_id]
            j = cnt_bin[pos_id]
            idx_pos = bin_to_idx[pos_id][step * j: step * (j+1)]

            step = point_per_bin[neg_id]
            j = cnt_bin[neg_id]
            idx_neg = bin_to_idx[neg_id][step * j: step * (j+1)]

            Xc = torch.cat([X[idx_pos], X[idx_neg]])
            yc = torch.cat([y[idx_pos], y[idx_neg]])

            cut_id = int(len(yc) * test_size)
            client_tr_assignments.append(FastDataLoader(
                Xc[:-cut_id],
                yc[:-cut_id],
                num_labels=2,
                batch_size=batch_size,
                shuffle=True,
                percentage=1.0
            ))

            client_te_assignments.append(FastDataLoader(
                Xc[-cut_id:],
                yc[-cut_id:],
                num_labels=2,
                batch_size=batch_size,
                shuffle=False,
                percentage=1.0
            ))

        return DummyDataContainer(client_tr_assignments, client_te_assignments, server_te, 2)


Datasets._DATASET_MAP = {
    "mnist": Datasets.MNIST,
    "svhn": Datasets.SVHN,
    "mnistm": Datasets.MNISTM,
    "femnist": Datasets.FEMNIST,
    "emnist": Datasets.EMNIST,
    "cifar10": Datasets.CIFAR10,
    "cifar100": Datasets.CIFAR100,
    "tiny_imagenet": Datasets.TINY_IMAGENET,
    "shakespeare": Datasets.SHAKESPEARE,
    "fashion_mnist": Datasets.FASHION_MNIST,
    "cinic10": Datasets.CINIC10,
    "fcube": Datasets.FCUBE
}



"""Implementation of the [FedSAM22]_ algorithm.

References:
    .. [FedSAM22] Caldarola, D., Caputo, B., & Ciccone, M. Improving Generalization in Federated
       Learning by Seeking Flat Minima. In ECCV (2022).
       URL: https://arxiv.org/abs/2203.11834
"""
import sys
from typing import Any, Iterable, Union

import torch
from torch.optim import Optimizer

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from . import CentralizedFL  # NOQA

# This implementation is based on
# https://github.com/bytedance/FedDecorr/blob/master/approach/fedsam.py

__all__ = [
    "SAMOptimizer",
    "FedSAMClient",
    "FedSAM"
]


class SAMOptimizer(torch.optim.Optimizer):

    def __init__(self,
                 params: Union[Iterable[torch.Tensor], Iterable[dict[str, Any]]],
                 base_optimizer: Optimizer = torch.optim.SGD,
                 rho: float = 0.05,
                 adaptive: bool = False,
                 **kwargs: dict[str, Any]):
        assert rho >= 0.0, f"Invalid rho, should be non-negative: {rho}"

        defaults = dict(rho=rho, adaptive=adaptive, **kwargs)
        super().__init__(params, defaults)
        self.defaults = defaults
        self.base_optimizer = base_optimizer(self.param_groups, **kwargs)
        self.param_groups = self.base_optimizer.param_groups

    @torch.no_grad()
    def first_step(self, zero_grad=False):
        grad_norm = self._grad_norm()
        for group in self.param_groups:
            scale = group["rho"] / (grad_norm + 1e-12)

            for p in group["params"]:
                if p.grad is None:
                    continue

                self.state[p]["old_p"] = p.data.clone()
                e_w = (torch.pow(p, 2) if group["adaptive"] else 1.0) * p.grad * scale.to(p)
                p.add_(e_w)  # climb to the local maximum "w + e(w)"

        if zero_grad:
            self.zero_grad()

    @torch.no_grad()
    def second_step(self, zero_grad=False):
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                p.data = self.state[p]["old_p"]  # get back to "w" from "w + e(w)"

        self.base_optimizer.step()  # do the actual "sharpness-aware" update

        if zero_grad:
            self.zero_grad()

    @torch.no_grad()
    def step(self, closure=None):
        assert closure is not None, "Sharpness Aware Minimization requires closure, \
            but it was not provided"
        closure = torch.enable_grad()(closure)  # the closure should do a full forward-backward pass

        loss1 = closure()
        self.first_step(zero_grad=True)
        loss2 = closure()
        self.second_step()

        return (loss1.item() + loss2.item()) / 2

    def _grad_norm(self):
        # put everything on the same device, in case of model parallelism
        shared_device = self.param_groups[0]["params"][0].device
        norm = torch.norm(
            torch.stack([
                        ((torch.abs(p) if group["adaptive"] else 1.0)
                         * p.grad).norm(p=2).to(shared_device)
                        for group in self.param_groups for p in group["params"]
                        if p.grad is not None
                        ]),
            p=2
        )
        return norm

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict)
        self.base_optimizer.param_groups = self.param_groups

    def __getstate__(self) -> dict[str, Any]:
        state = super().__getstate__()
        state["base_optimizer"] = self.base_optimizer
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        super().__setstate__(state)
        self.base_optimizer = state["base_optimizer"]


class FedSAMClient(Client):

    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,  # ignored
                 local_epochs: int,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 rho: float = 0.05,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(rho=rho)

    def _get_closure(self, X: torch.Tensor, y: torch.Tensor):
        def closure():
            y_hat = self.model(X)
            loss = self.hyper_params.loss_fn(y_hat, y)
            loss.backward()
            self._clip_grads(self.model)
            return loss
        return closure

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)
        self.model.to(self.device)
        self.model.train()
        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model,
                                                                 rho=self.hyper_params.rho)
        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                running_loss += self.optimizer.step(self._get_closure(X, y))

            self.scheduler.step()

        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        clear_cuda_cache()
        return running_loss


class FedSAM(CentralizedFL):

    def can_override_optimizer(self) -> bool:
        return False

    def get_optimizer_class(self) -> Optimizer:
        return SAMOptimizer

    def get_client_class(self) -> Client:
        return FedSAMClient



from __future__ import annotations

import sys

import torch

sys.path.append(".")
sys.path.append("..")


# 'EncoderHeadNet',
# 'GlobalLocalNet',
# 'HeadGlobalEncoderLocalNet',
# 'EncoderGlobalHeadLocalNet',


from fluke.nets import EncoderGlobalHeadLocalNet  # NOQA
from fluke.nets import (FEMNIST_CNN, MNIST_2NN, MNIST_CNN, MNIST_LR,  # NOQA
                        VGG9, CifarConv2, FedAVGCNN, FedBN_CNN,
                        HeadGlobalEncoderLocalNet, LeNet5, MoonCNN, ResNet9,
                        ResNet18, ResNet18GN, ResNet34, ResNet50,
                        Shakespeare_LSTM)


def test_mnist_2nn():

    model = MNIST_2NN()

    x = torch.randn(1, 28, 28)
    y = model(x)

    y1 = model.forward(x)
    z = model.forward_encoder(x)
    y2 = model.forward_head(z)

    assert z.shape == (1, 100)
    assert torch.allclose(y, y1)
    assert torch.allclose(y, y2)

    model = MNIST_2NN(softmax=True)

    y = model(x)
    assert y.shape == (1, 10)
    assert torch.allclose(y.sum(), torch.tensor(1.0))


def test_mnist_lr():
    model = MNIST_LR()
    x = torch.randn(1, 28, 28)
    y = model(x)

    assert y.shape == (1, 10)
    assert torch.allclose(y.sum(), torch.tensor(1.0))


def test_convnets():
    model = MNIST_CNN()
    x = torch.randn(1, 1, 28, 28)
    y1 = model(x)
    z = model.forward_encoder(x)
    y2 = model.forward_head(z)
    assert y1.shape == (1, 10)
    assert z.shape == (1, 1024)
    assert torch.allclose(y1, y2)

    model = FEMNIST_CNN()
    x = torch.randn(1, 1, 28, 28)
    y1 = model(x)
    z = model.forward_encoder(x)
    y2 = model.forward_head(z)
    assert y1.shape == (1, 62)
    assert z.shape == (1, 3136)
    assert torch.allclose(y1, y2)

    model = VGG9()
    x = torch.randn(1, 1, 28, 28)
    y1 = model(x)
    z = model.forward_encoder(x)
    y2 = model.forward_head(z)
    assert y1.shape == (1, 62)
    assert z.shape == (1, 512)
    assert torch.allclose(y1, y2)

    model = FedAVGCNN()
    x = torch.randn(1, 3, 32, 32)
    y1 = model(x)
    z = model.forward_encoder(x)
    y2 = model.forward_head(z)
    assert y1.shape == (1, 10)
    assert z.shape == (1, 4096)
    assert torch.allclose(y1, y2)

    model = LeNet5(10)
    x = torch.randn(1, 3, 32, 32)
    z = model.forward_encoder(x)
    y1 = model(x)
    y2 = model.forward_head(z)
    assert y1.shape == (1, 10)
    assert z.shape == (1, 400)
    assert torch.allclose(y1, y2)

    model = MoonCNN()
    x = torch.randn(1, 3, 32, 32)
    z = model.forward_encoder(x)
    y1 = model(x)
    y2 = model.forward_head(z)
    assert y1.shape == (1, 10)
    assert z.shape == (1, 84)
    assert torch.allclose(y1, y2)

    model = FedBN_CNN()
    x = torch.randn(2, 1, 28, 28)
    z = model.forward_encoder(x)
    y1 = model(x)
    y2 = model.forward_head(z)
    assert y1.shape == (2, 10)
    assert z.shape == (2, 6272)
    assert torch.allclose(y1, y2)

    model = CifarConv2()
    x = torch.randn(2, 3, 32, 32)
    z = model.forward_encoder(x)
    y2 = model.forward_head(z)
    y1 = model(x)
    assert y1.shape == (2, 10)
    assert z.shape == (2, 1600)
    assert torch.allclose(y1, y2)

    model = ResNet9()
    x = torch.randn(1, 3, 32, 32)
    z = model.forward_encoder(x)
    y1 = model(x)
    y2 = model.forward_head(z)
    assert y1.shape == (1, 100)
    assert z.shape == (1, 1024)
    assert torch.allclose(y1, y2)

    model = ResNet18()
    x = torch.randn(2, 3, 32, 32)
    y1 = model(x)
    assert y1.shape == (2, 10)

    model = ResNet18GN()
    x = torch.randn(2, 3, 32, 32)
    y1 = model(x)
    assert y1.shape == (2, 10)

    model = ResNet34()
    x = torch.randn(2, 3, 32, 32)
    y1 = model(x)
    assert y1.shape == (2, 100)

    model = ResNet50()
    x = torch.randn(2, 3, 32, 32)
    y1 = model(x)
    assert y1.shape == (2, 100)


def test_recurrent():
    model = Shakespeare_LSTM()
    x = torch.randint(0, 100, (1, 10))
    z = model.forward_encoder(x)
    y1 = model(x)
    y2 = model.forward_head(z)
    assert z.shape == (1, 256)
    assert y1.shape == (1, 100)
    assert y2.shape == (1, 100)
    assert torch.allclose(y1, y2)


def test_global_local():
    nn = MNIST_2NN()
    model = HeadGlobalEncoderLocalNet(nn)
    assert nn.encoder is model.get_local()
    assert nn.head is model.get_global()

    model = EncoderGlobalHeadLocalNet(nn)
    assert nn.encoder is model.get_global()
    assert nn.head is model.get_local()


if __name__ == "__main__":
    test_mnist_2nn()
    test_mnist_lr()
    test_convnets()
    test_recurrent()
    test_global_local()
    # 99% coverage on nets.py



"""Implementation of the FedBN [FedBN21]_ algorithm.

References:
    .. [FedBN21] Xiaoxiao Li, Meirui JIANG, Xiaofei Zhang, Michael Kamp, and Qi Dou. FedBN:
       Federated Learning on Non-IID Features via Local Batch Normalization. In ICLR (2021).
       URL: https://openreview.net/pdf?id=6YEQUn0QICG
"""
import sys

import torch

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA

# Same idea in https://link.springer.com/chapter/10.1007/978-3-030-60548-3_13

__all__ = [
    "FedBNClient",
    "FedBN"
]


class FedBNClient(Client):

    def _get_bn_keys(self, module: torch.nn.Module, running_key: str = "") -> list:
        keys = []
        for key, value in module.named_children():
            if isinstance(value, torch.nn.modules.batchnorm._BatchNorm):
                keys.append(f"{running_key}{key}")
            else:
                keys += self._get_bn_keys(value, running_key=key + ".")
        return keys

    def receive_model(self) -> None:
        global_model = self.channel.receive(self, self.server, msg_type="model").payload
        if self.model is None:
            self.model = global_model

            to_skip = self._get_bn_keys(self.model)
            self._to_keep = []
            for key in self.model.state_dict():
                ok = True
                for skip in to_skip:
                    if key.startswith(skip):
                        ok = False
                        break
                if ok:
                    self._to_keep.append(key)

        else:
            with torch.no_grad():
                for key in global_model.state_dict().keys():
                    if key in self._to_keep:
                        self.model.state_dict()[key].data.copy_(
                            global_model.state_dict()[key].clone())


class FedBN(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedBNClient



"""This module contains (as submodules) the implementation of several the federated learning
algorithms."""
from __future__ import annotations

import os
import sys
import uuid
import warnings
from copy import deepcopy
from typing import Any, Iterable, Union

import torch

sys.path.append(".")
sys.path.append("..")

from .. import DDict, FlukeENV  # NOQA
from ..client import Client, PFLClient  # NOQA
from ..comm import ChannelObserver  # NOQA
from ..data import DataSplitter, FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import (ClientObserver, OptimizerConfigurator,  # NOQA
                     ServerObserver, get_loss, get_model)

__all__ = [
    'CentralizedFL',
    'PersonalizedFL',
    'apfl',
    'ccvr',
    'ditto',
    'dpfedavg',
    'fat',
    'fedala',
    'fedamp',
    'fedavg',
    'fedavgm',
    'fedaws',
    'fedbabu',
    'fedbn',
    'feddyn',
    'fedexp',
    'fedhp',
    'fedlc',
    'fedld',
    'fednh',
    'fednova',
    'fedopt',
    'fedper',
    'fedproto',
    'fedprox',
    'fedrep',
    'fedrod',
    'fedrs',
    'fedsam',
    'fedsgd',
    'gear',
    'kafe',
    'lg_fedavg',
    'moon',
    'per_fedavg',
    'pfedme',
    'scaffold',
    'superfed'
]


class CentralizedFL(ServerObserver):
    """Centralized Federated Learning algorithm.
    This class is a generic implementation of a centralized federated learning algorithm that
    follows the Federated Averaging workflow. This class represents the entry point to the
    federated learning algorithm.
    Each new algorithm should inherit from this class and implement the specific logic of the
    algorithm. The main components of the algorithm are:

    - ``Clients``: Each client should implement :class:`fluke.client.Client` class and the specific
      specialization must be defined in the :meth:`get_client_class` method. The initialization of
      the clients is done in the :meth:`init_clients` method.
    - ``Server``: The server is the entity that coordinates the training process. It should
      implement the :class:`fluke.server.Server` class and the specific specialization must be
      defined in the :meth:`get_server_class` method. The initialization of the server is done in
      the :meth:`init_server` method.
    - ``Optimizer``: The optimizer used by the clients. The default optimizer class is defined in
      the :meth:`get_optimizer_class` method.

    To run the algorithm, the :meth:`run` method should be called with the number of rounds and the
    percentage of eligible clients. This method will call the :meth:`Server.fit` method which will
    orchestrate the training process.

    Args:
        n_clients (int): Number of clients.
        data_splitter (DataSplitter): Data splitter object.
        hyper_params (DDict): Hyperparameters of the algorithm. This set of hyperparameteers should
          be divided in two parts: the client hyperparameters and the server hyperparameters.

    See Also:
        - :class:`fluke.client.Client`
        - :class:`fluke.server.Server`
        - :class:`PersonalizedFL`
    """

    def __init__(self,
                 n_clients: int,
                 data_splitter: DataSplitter,
                 hyper_params: DDict | dict[str, Any],
                 **kwargs: dict[str, Any]):
        self._id = str(uuid.uuid4().hex)
        FlukeENV().open_cache(self._id)
        self.hyper_params = hyper_params if isinstance(hyper_params, DDict) else DDict(hyper_params)
        self.n_clients = n_clients
        (clients_tr_data, clients_te_data), server_data = \
            data_splitter.assign(n_clients, hyper_params.client.batch_size)
        # Federated model
        model = get_model(mname=hyper_params.model,
                          **hyper_params.net_args if 'net_args' in hyper_params else {}
                          ) if isinstance(hyper_params.model, str) else hyper_params.model
        self.clients = None
        self.server = None
        self.init_clients(clients_tr_data, clients_te_data, hyper_params.client)
        self.init_server(model, server_data, hyper_params.server)

        for client in self.clients:
            client.set_server(self.server)

    @property
    def id(self) -> str:
        """Get the unique identifier of this instance of the algorithm.

        Returns:
            str: Unique identifier of the instance of the algorithm.
        """
        return str(self._id)

    def can_override_optimizer(self) -> bool:
        """Return whether the optimizer can be changed user-side.
        Generally, the optimizer can be configured by the user. However, in some cases, the
        algorithm may require a specific optimizer and the user should not be able to change it.

        Returns:
            bool: Whether the optimizer can be changed user-side.
        """
        return True

    def get_optimizer_class(self) -> torch.optim.Optimizer:
        """Get the optimizer class.

        Returns:
            torch.optim.Optimizer: Optimizer class.
        """
        return torch.optim.SGD

    def get_client_class(self) -> Client:
        """Get the client class.
        This method should be overriden by the subclasses when a different client class is defined.
        This allows to reuse all the logic of the algorithm and only change the client class.

        Returns:
            Client: Client class.
        """
        return Client

    def get_server_class(self) -> Server:
        """Get the server class.
        This method should be overriden by the subclasses when a different server class is defined.
        This allows to reuse all the logic of the algorithm and only change the server class.

        Returns:
            Server: Server class.
        """
        return Server

    def _fix_opt_cfg(self, cfg_opt: DDict) -> None:

        if "name" not in cfg_opt:
            cfg_opt.name = "SGD"

        if not self.can_override_optimizer() and \
                (cfg_opt.name != self.get_optimizer_class().__name__ or
                 not isinstance(cfg_opt.name, self.get_optimizer_class())):
            old_name = cfg_opt.name if isinstance(cfg_opt.name, str) else cfg_opt.name.__name__
            warnings.warn(f"The algorithm does not support the optimizer {old_name}. "
                          f"Using {self.get_optimizer_class().__name__} instead.")
            cfg_opt.name = self.get_optimizer_class()

    def init_clients(self,
                     clients_tr_data: list[FastDataLoader],
                     clients_te_data: list[FastDataLoader],
                     config: DDict) -> None:
        """Initialize the clients.

        Args:
            clients_tr_data (list[FastDataLoader]): List of training data loaders, one for
               each client.
            clients_te_data (list[FastDataLoader]): List of test data loaders, one for
               each client. The test data loaders can be ``None``.
            config (DDict): Configuration of the clients.

        Important:
            For more deatils about the configuration of the clients, see the
            :ref:`configuration <configuration>` page.

        See Also:
            :class:`fluke.client.Client`
        """

        self._fix_opt_cfg(config.optimizer)
        optimizer_cfg = OptimizerConfigurator(optimizer_cfg=config.optimizer,
                                              scheduler_cfg=config.scheduler)
        self.loss = get_loss(config.loss) if isinstance(config.loss, str) else config.loss()
        self.clients = [
            self.get_client_class()(
                index=i,
                train_set=clients_tr_data[i],
                test_set=clients_te_data[i],
                optimizer_cfg=optimizer_cfg,
                loss_fn=deepcopy(self.loss),
                **config.exclude('optimizer', 'loss', 'batch_size', 'scheduler')
            )
            for i in range(self.n_clients)]

    def init_server(self, model: Any, data: FastDataLoader, config: DDict):
        """Initailize the server.

        Args:
            model (Any): The global model.
            data (FastDataLoader): The server-side test set.
            config (DDict): Configuration of the server.
        """
        self.server: Server = self.get_server_class()(model, data, self.clients, **config)
        if FlukeENV().get_save_options()[0] is not None:
            self.server.attach(self)

    def set_callbacks(self, callbacks: Union[callable, Iterable[callable]]):
        """Set the callbacks for the server the clients and the channel.

        The callbacks are expected to be instances of the :class:`fluke.server.ServerObserver`,
        :class:`fluke.client.ClientObserver` or :class:`fluke.comm.ChannelObserver` classes.
        Each callback will be attached to the corresponding entity.

        Args:
            callbacks (Union[callable, Iterable[callable]]): Callbacks to attach to the algorithm.
        """
        if not isinstance(callbacks, Iterable):
            callbacks = [callbacks]
        self.server.attach([c for c in callbacks if isinstance(c, ServerObserver)])
        self.server.channel.attach([c for c in callbacks if isinstance(c, ChannelObserver)])
        for client in self.clients:
            client.attach([c for c in callbacks if isinstance(c, ClientObserver)])

    def run(self,
            n_rounds: int,
            eligible_perc: float,
            finalize: bool = True,
            **kwargs: dict[str, Any]):
        """Run the federated algorithm.
        This method will call the :meth:`Server.fit` method which will orchestrate the training
        process.

        Args:
            n_rounds (int): Number of rounds.
            eligible_perc (float): Percentage of eligible clients.
            finalize (bool, optional): Whether to finalize the training process.
              Defaults to ``True``.
            **kwargs (dict[str, Any]): Additional keyword arguments.
        """
        self.server.fit(n_rounds=n_rounds, eligible_perc=eligible_perc, finalize=finalize, **kwargs)

    def __str__(self) -> str:
        algo_hp = ",\n\t".join(
            [f"{h}={str(v)}"
             for h, v in self.hyper_params.items() if h not in ['client', 'server']]
        )
        algo_hp = f"\n\t{algo_hp}," if algo_hp else ""

        if self.clients is None:
            client_str = "Client?"
        else:
            client_str = str(self.clients[0]).replace("[0]", f"[0-{self.n_clients-1}]")

        if self.server is None:
            server_str = "Server?"
        else:
            server_str = str(self.server)

        return f"{self.__class__.__name__}[{self._id}]" + \
            f"({algo_hp}\n\t{client_str},\n\t{server_str}\n)"

    def __repr__(self) -> str:
        return str(self)

    def save(self, path: str, global_only: bool = False, round: int | None = None) -> str:
        """Save the algorithm state into files in the specified directory.

        Note:
            To avoid overwriting previous saved states, the folder name will be suffixed with the
            unique (randomly generated) identifier of the algorithm.

        Args:
            path (str): Path to the folder where the algorithm state will be saved.
            global_only (bool, optional): Whether to save only the global model. Defaults to
                ``False``.
            round (int, optional): Round number. Defaults to ``None``.

        Returns:
            str: Path to the folder where the algorithm state was saved.
        """
        path = f"{path}_{self._id}"
        if not os.path.exists(path):
            os.makedirs(path)

        prefix = f"r{str(round).zfill(4)}_" if round is not None else ""
        self.server.save(os.path.join(path, f"{prefix}server.pth"))
        if not global_only:
            for i, client in enumerate(self.clients):
                client.save(os.path.join(path, f"{prefix}client_{i}.pth"))

        return path

    def load(self, path: str, round: int | None = None) -> None:
        """Load the algorithm state from the specified folder

        Args:
            path (str): Path to the folder where the algorithm state is saved.
            round (int, optional): Round number. Defaults to ``None``.
        """
        if round is not None:
            prefix = f"r{str(round).zfill(4)}_"
        else:
            prefix = ""
            # search in path for the last round
            max_round = -1
            for f in os.listdir(path):
                if f.startswith("r") and f.endswith("_server.pth"):
                    round = int(f[1:5])
                    max_round = max(max_round, round)
            if max_round != -1:
                prefix = f"r{str(max_round).zfill(4)}_"

        self.server.load(os.path.join(path, f"{prefix}server.pth"))
        for i, client in enumerate(self.clients):
            client.load(os.path.join(path, f"{prefix}client_{i}.pth"), deepcopy(self.server.model))

    # ServerObserver methods
    def end_round(self, round: int) -> None:
        path, freq, g_only = FlukeENV().get_save_options()
        if freq > 0 and round % freq == 0:
            self.save(path, g_only, round)

    # ServerObserver methods
    def finished(self, round: int) -> None:
        path, freq, g_only = FlukeENV().get_save_options()
        if freq == -1:
            self.save(path, g_only, round-1)


class PersonalizedFL(CentralizedFL):
    """Personalized Federated Learning algorithm. This class is a simple extension of the
    :class:`CentralizedFL` class where the clients are expected to implement the
    :class:`fluke.client.PFLClient` class (see :meth:`get_client_class`). The main difference with
    respect to the :class:`CentralizedFL` class is that the clients initialization requires a model
    that is the personalized model of the client.

    Important:
        Differently from :class:`CentralizedFL`, which is actually the FedAvg algorithm, the
        :class:`PersonalizedFL` class must not be used as is because it is a generic implementation
        of a personalized federated learning algorithm. The subclasses of this class should
        implement the specific logic of the personalized federated learning algorithm.

    See Also:
        - :class:`CentralizedFL`
        - :class:`fluke.client.PFLClient`
    """

    def init_clients(self,
                     clients_tr_data: list[FastDataLoader],
                     clients_te_data: list[FastDataLoader],
                     config: DDict) -> None:

        if isinstance(config.model, str):
            model = get_model(mname=config.model, **config.net_args if 'net_args' in config else {})
        elif isinstance(config.model, torch.nn.Module):
            model = config.model
        else:
            raise ValueError("Invalid model configuration. \
                             It should be a string or a torch.nn.Module")

        self._fix_opt_cfg(config.optimizer)
        optimizer_cfg = OptimizerConfigurator(optimizer_cfg=config.optimizer,
                                              scheduler_cfg=config.scheduler)
        self.loss = get_loss(config.loss) if isinstance(config.loss, str) else config.loss()
        self.clients = [
            self.get_client_class()(
                index=i,
                model=deepcopy(model),
                train_set=clients_tr_data[i],
                test_set=clients_te_data[i],
                optimizer_cfg=optimizer_cfg,
                loss_fn=deepcopy(self.loss),
                **config.exclude('optimizer', 'loss', 'batch_size', 'model', 'scheduler')
            )
            for i in range(self.n_clients)]



"""Implementation of the [FedOpt21]_ algorithm.

References:
    .. [FedOpt21] Sashank Reddi, Zachary Charles, Manzil Zaheer, Zachary Garrett, Keith Rush,
       Jakub Konečný, Sanjiv Kumar, H. Brendan McMahan. Adaptive Federated Optimization.
       In ICLR (2021). URL: https://openreview.net/pdf?id=LkFG3lB13U5
"""
import sys
from collections import OrderedDict
from copy import deepcopy
from typing import Iterable

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils.model import get_trainable_keys  # NOQA

__all__ = [
    "FedOptServer",
    "FedOpt"
]


class FedOptServer(Server):
    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 mode: str = "adam",
                 lr: float = 0.001,
                 beta1: float = 0.9,
                 beta2: float = 0.999,
                 tau: float = 0.0001,
                 weighted: bool = True):
        super().__init__(model=model, test_set=test_set, clients=clients, weighted=weighted)
        assert mode in {"adam", "yogi", "adagrad"}, \
            "'mode' must be one of {'adam', 'yogi', 'adagrad'}"
        assert 0 <= beta1 < 1, "beta1 must be in [0, 1)"
        assert 0 <= beta2 < 1, "beta2 must be in [0, 1)"

        self.hyper_params.update(
            mode=mode,
            lr=lr,
            beta1=beta1,
            beta2=beta2,
            tau=tau
        )
        self._init_moments()

    def _init_moments(self):
        self.m_t = OrderedDict()
        self.v_t = OrderedDict()
        for key in self.model.state_dict().keys():
            if "num_batches_tracked" not in key:
                self.m_t[key] = torch.zeros_like(self.model.state_dict()[key])
                # This guarantees that the second moment is >= 0 and >= tau^2
                self.v_t[key] = torch.zeros_like(self.model.state_dict()[key])

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        prev_model = deepcopy(self.model)
        super().aggregate(eligible, client_models)
        aggregated = self.model.state_dict()
        server_sd = prev_model.state_dict()
        b1, b2 = self.hyper_params.beta1, self.hyper_params.beta2
        eta, tau = self.hyper_params.lr, self.hyper_params.tau

        trainable_keys = get_trainable_keys(self.model)
        d_t = {k: aggregated[k] - server_sd[k] for k in trainable_keys}
        self.m_t = {k: b1 * self.m_t[k] + (1 - b1) * d_t[k] for k in trainable_keys}

        if self.hyper_params.mode == "adam":
            self.v_t = {k: b2 * self.v_t[k] + (1 - b2) * d_t[k] ** 2 for k in trainable_keys}
        elif self.hyper_params.mode == "yogi":
            self.v_t = {k: self.v_t[k] - (1 - b2) * (d_t[k] ** 2) *
                        torch.sign(self.v_t[k] - d_t[k] ** 2) for k in trainable_keys}
        elif self.hyper_params.mode == "adagrad":
            self.v_t = {k: self.v_t[k] + d_t[k] ** 2 for k in trainable_keys}
        else:
            raise ValueError(f"Unknown mode: {self.hyper_params.mode}")

        update = {k: eta * self.m_t[k] / (torch.sqrt(self.v_t[k]) + tau) for k in trainable_keys}
        agg_model_sd = {k: server_sd[k] + update[k] for k in trainable_keys}
        self.model.load_state_dict(agg_model_sd, strict=False)


class FedOpt(CentralizedFL):

    def get_server_class(self) -> Server:
        return FedOptServer



from __future__ import annotations

import sys
import warnings

import numpy as np
import pytest
import torch

sys.path.append(".")
sys.path.append("..")

from fluke import FlukeENV, FlukeCache  # NOQA
from fluke.comm import Channel, ChannelObserver, Message  # NOQA


def test_message():
    msg = Message(payload="a", msg_type="type_test", sender=None)
    assert msg.msg_type == "type_test"
    assert msg.payload == "a"
    assert msg.sender is None
    assert msg.size == 1
    msg = Message(payload=None, msg_type="type_test", sender=None)
    assert msg.size == 1
    msg = Message(payload=[1, 2, 3, "abc"], msg_type="type_test", sender=None)
    assert msg.size == 6
    msg = Message(payload={"a": 1, "b": 2}, msg_type="type_test", sender=None)
    assert msg.size == 4
    msg = Message(payload=17, msg_type="type_test", sender=None)
    assert msg.size == 1
    msg = Message(payload=17.0, msg_type="type_test", sender=None)
    assert msg.size == 1
    msg = Message(payload=True, msg_type="type_test", sender=None)
    assert msg.size == 1
    msg = Message(payload=np.array([1, 2, 3]), msg_type="type_test", sender=None)
    assert msg.size == 3
    msg = Message(payload=np.array([[1, 2], [3, 4]]), msg_type="type_test", sender=None)
    assert msg.size == 4
    msg = Message(payload=torch.FloatTensor([1, 2, 3]), msg_type="type_test", sender=None)
    assert msg.size == 3
    msg = Message(payload=torch.nn.Linear(10, 10), msg_type="type_test", sender=None)
    assert msg.size == 100 + 10
    msg = Message(payload=FlukeCache._ObjectRef(), msg_type="type_test", sender=None)
    assert msg.size == 0

    class A():
        pass

    warnings.filterwarnings("ignore")
    msg = Message(payload=A(), msg_type="type_test", sender=None)
    assert msg.size == 0

    msg2 = Message(payload="a", msg_type="type_test", sender=None)
    assert msg.id != msg2.id

    msg = Message(payload="prova", msg_type="type_test", sender=None)
    assert str(
        msg) == f"Message[{msg.id}](type=type_test, from=None, payload=prova, size=5, inmemory=True)"
    assert msg.__repr__(
    ) == f"Message[{msg.id}](type=type_test, from=None, payload=prova, size=5, inmemory=True)"

    tensor = torch.FloatTensor([1, 2, 3])
    FlukeENV().set_inmemory(False)
    FlukeENV().open_cache("tmp_comm")
    msg = Message(tensor, "type_test", "sender", inmemory=False)
    assert FlukeENV().get_cache().occupied == 1
    assert not msg.cache()
    msg.ram()
    assert FlukeENV().get_cache().occupied == 0
    assert msg.cache()
    assert FlukeENV().get_cache().occupied == 1
    assert torch.all(msg.payload == tensor)
    assert FlukeENV().get_cache().occupied == 0
    FlukeENV().get_cache().cleanup()
    FlukeENV().close_cache()


def test_channel():

    chobs = ChannelObserver()
    chobs.message_received(Message("a", "type_test", "sender"))
    assert isinstance(chobs, ChannelObserver)

    class Observer(ChannelObserver):
        def message_received(self, msg):
            self.msg = msg

    FlukeENV().open_cache("tmp_comm")
    channel = Channel()
    assert not channel.buffer
    obs = Observer()
    channel.attach(obs)
    msg = Message("a", "type_test", "sender")
    channel.send(msg, "pippo")
    assert len(channel.buffer) == 1
    assert "pippo" in channel._buffer
    assert len(channel["pippo"]) == 1
    assert channel["pippo"][0] == msg
    msg = channel.receive("pippo")
    assert msg.payload == "a"
    assert msg.msg_type == "type_test"
    assert msg.sender == "sender"
    assert not channel.buffer["pippo"]
    channel.send(Message("a", "type_test", "sender"), "pippo")
    channel.clear("pippo")
    assert not channel.buffer["pippo"]
    assert obs.msg == msg

    channel.broadcast(Message("a", "type_test", "sender"), ["pippo", "pluto"])
    assert len(channel._buffer) == 2
    assert "pippo" in channel.buffer
    assert "pluto" in channel.buffer

    channel.send(Message("b", "type_test", "sender"), "pippo")
    msg = channel.receive("pippo", "sender", "type_test")
    assert msg.payload == "a"
    msg = channel.receive("pippo", "sender", "type_test")
    assert msg.payload == "b"

    with pytest.raises(ValueError):
        channel.receive("pippo", "sender", "type_test")

    FlukeENV().close_cache()


if __name__ == "__main__":
    test_message()
    test_channel()
    # 99% coverage comm.py



"""Implementation of the [FedLC22]_ algorithm.

References:
    .. [FedLC22] Jie Zhang, Zhiqi Li, Bo Li, Jianghe Xu, Shuang Wu, Shouhong Ding, Chao Wu.
       Federated Learning with Label Distribution Skew via Logits Calibration. In ICML (2022).
       URL: https://arxiv.org/abs/2209.00189
"""
import sys
from typing import Any, Literal

import numpy as np
import torch

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import OptimizerConfigurator  # NOQA
from . import CentralizedFL  # NOQA


__all__ = [
    "CalibratedLoss",
    "FedLCClient",
    "FedLC"
]


class CalibratedLoss(torch.nn.Module):
    """Calibrated Loss function.

    Args:
        tau (float): calibration parameter.
        label_distrib (torch.Tensor): Label distribution.
    """

    def __init__(self,
                 tau: float,
                 label_distrib: torch.Tensor,
                 reduction: Literal["mean", "sum"] = "mean"):
        super().__init__()
        self.tau = tau
        self.label_distrib = label_distrib
        self.reduction = reduction

    def forward(self, logit: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        out = logit - self.tau * self.label_distrib.to(logit.device)**(-0.25)
        return torch.nn.functional.cross_entropy(out, y, reduction=self.reduction)

    def __str__(self):
        return f"CalibratedLoss(tau={self.tau}, reduction={self.reduction})"

    def __repr__(self):
        return str(self)


class FedLCClient(Client):

    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,  # not used
                 local_epochs: int,
                 tau: float,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        label_counter = torch.zeros(train_set.num_labels)
        uniq_val, uniq_count = np.unique(train_set.tensors[1], return_counts=True)
        for i, c in enumerate(uniq_val.tolist()):
            label_counter[c] = max(1e-8, uniq_count[i])
        label_counter = label_counter.unsqueeze(dim=0)  # .to(self.device)
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=CalibratedLoss(tau, label_counter),
                         local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(tau=tau)


class FedLC(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedLCClient



"""Implementation of the [FedRS21]_ algorithm.

References:
    .. [FedRS21] Xin-Chun Li and De-Chuan Zhan. FedRS: Federated Learning with Restricted
       Softmax for Label Distribution Non-IID Data. In KDD (2021).
       URL: https://doi.org/10.1145/3447548.3467254
"""
import sys
from typing import Any, Literal

import numpy as np
import torch

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import OptimizerConfigurator  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "RSLoss",
    "FedRSClient",
    "FedRS"
]


class RSLoss(torch.nn.Module):
    """Restricted Softmax Loss function.

    Args:
        class_scaling (torch.Tensor): Class scaling factor.
        reduction (Literal["mean", "sum"]): Specifies the reduction to apply to the output.

    See Also:
        This loss function is very similar to the one used in ``FedLC``, i.e.,
        :class:`fluke.algorithms.fedlc.CalibratedLoss`.
    """

    def __init__(self, class_scaling: torch.Tensor, reduction: Literal["mean", "sum"] = "mean"):
        super().__init__()
        self.class_scaling = class_scaling
        self.reduction = reduction

    def forward(self, logit: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        out = logit * self.class_scaling.to(logit.device)
        return torch.nn.functional.cross_entropy(out, y, reduction=self.reduction)

    def __str__(self):
        return "RSLoss()"

    def __repr__(self):
        return str(self)


class FedRSClient(Client):

    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,  # ignored
                 local_epochs: int,
                 alpha: float,
                 count_as_missing: int = 2,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        uniq_val, uniq_cnt = np.unique(train_set.tensors[1], return_counts=True)
        class_scaling = torch.ones(train_set.num_labels) * alpha
        for i, c in enumerate(uniq_val):
            if uniq_cnt[i] > count_as_missing:
                class_scaling[c] = 1.0
        class_scaling = class_scaling.unsqueeze(dim=0)
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=RSLoss(class_scaling),
                         local_epochs=local_epochs, fine_tuning_epochs=fine_tuning_epochs,
                         clipping=clipping, **kwargs)
        self.hyper_params.update(alpha=alpha, count_as_missing=count_as_missing)


class FedRS(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedRSClient
